#include <node.h>
#include <iostream>
#include "Matrix.h"
#include "Vector.h"
#include "LogisticMFEngine.h"
#include "ALSEngine.h"
#include "Metrics.h"

namespace demo {

using v8::FunctionCallbackInfo;
using v8::Isolate;
using v8::Local;
using v8::Object;
using v8::String;
using v8::Value;
using v8::Array;

void testLogisticMf(const FunctionCallbackInfo<Value>& args) {
  Isolate* isolate = args.GetIsolate();

  Matrix ratingMatrix(Local<Array>::Cast(args[0]));
  Matrix userVectors(Local<Array>::Cast(args[1]));
  Matrix itemVectors(Local<Array>::Cast(args[2]));
  Vector userBiases(Local<Array>::Cast(args[3]));
  Vector itemBiases(Local<Array>::Cast(args[4]));

  double regularization = args[5]->NumberValue();
  double gamma = args[6]->NumberValue();
  size_t iterations = args[7]->IntegerValue();
  size_t factors = args[8]->IntegerValue();

  size_t numUsers = ratingMatrix.rows();
  size_t numItems = ratingMatrix.cols();
  size_t numZeros = numUsers * numItems;
  double total = 0;

  for(size_t i = 0; i < numUsers; i++) {
    for(size_t j = 0; j < numItems; j++) {
      double value = ratingMatrix(i, j);
      if(value == 0) continue;
      numZeros -= 1;
      total += value;
    }
  }

  double alpha = numZeros / total;
  Matrix newRatingMatrix = ratingMatrix * alpha;

  LogisticMFEngine engine(
    newRatingMatrix,
    userVectors, itemVectors,
    userBiases, itemBiases,
    regularization, gamma,
    iterations, factors
  );

  Matrix X = engine.train();

  args.GetReturnValue().Set(X.convertToLocalArray(isolate));
}

void logisticMF(const FunctionCallbackInfo<Value>& args) {
  Isolate* isolate = args.GetIsolate();

  Matrix ratingMatrix(Local<Array>::Cast(args[0]));
  double regularization = args[1]->NumberValue();
  double gamma = args[2]->NumberValue();
  size_t factors = args[3]->IntegerValue();
  size_t iterations = args[4]->IntegerValue();
  size_t seed = args[5]->IntegerValue();

  size_t numUsers = ratingMatrix.rows();
  size_t numItems = ratingMatrix.cols();
  size_t numZeros = numUsers * numItems;
  double total = 0;

  for(size_t i = 0; i < numUsers; i++) {
    for(size_t j = 0; j < numItems; j++) {
      double value = ratingMatrix(i, j);
      if(value == 0) continue;
      numZeros -= 1;
      total += value;
    }
  }

  double alpha = numZeros / total;
  Matrix newRatingMatrix = ratingMatrix * alpha;

  LogisticMFEngine engine(
    newRatingMatrix,
    factors, iterations,
    regularization, gamma,
    seed
  );

  Matrix X = engine.train();

  args.GetReturnValue().Set(X.convertToLocalArray(isolate));
}

void testAls(const FunctionCallbackInfo<Value>& args) {
  Isolate* isolate = args.GetIsolate();

  Matrix ratingMatrix(Local<Array>::Cast(args[0]));
  Matrix userFactors(Local<Array>::Cast(args[1]));
  Matrix itemFactors(Local<Array>::Cast(args[2]));

  double regularization = args[3]->NumberValue();
  size_t iterations = args[4]->IntegerValue();

  ALSEngine engine(
    ratingMatrix,
    userFactors,
    itemFactors,
    regularization,
    iterations
  );

  Matrix X = engine.train();
  args.GetReturnValue().Set(X.convertToLocalArray(isolate));
}

void als(const FunctionCallbackInfo<Value>& args) {
  Isolate* isolate = args.GetIsolate();

  Matrix ratingMatrix(Local<Array>::Cast(args[0]));
  double alpha = args[1]->NumberValue();
  size_t factors = args[2]->IntegerValue();
  double regularization = args[3]->NumberValue();
  size_t iterations = args[4]->IntegerValue();
  size_t seed = args[5]->IntegerValue();

  ALSEngine engine(
    ratingMatrix,
    alpha,
    regularization,
    factors,
    iterations,
    seed
  );

  Matrix X = engine.train();
  args.GetReturnValue().Set(X.convertToLocalArray(isolate));
}

void auc(const FunctionCallbackInfo<Value>& args) {

  Local<Array> labels = Local<Array>::Cast(args[0]);
  Local<Array> scores = Local<Array>::Cast(args[1]);

  size_t length = labels->Length();

  std::vector<double> newLabels;
  std::vector<double> newScores;

  for(size_t i = 0; i < length; i++) {
    newLabels.push_back( labels->Get(i)->NumberValue() );
    newScores.push_back( scores->Get(i)->NumberValue() );
  }

  AUC aucMetric;
  double aucScore = aucMetric.compute(newLabels, newScores);

  args.GetReturnValue().Set(aucScore);
}

void init(Local<Object> exports) {

  // Metrics
  NODE_SET_METHOD(exports, "auc", auc);

  // Production Algorithms
  NODE_SET_METHOD(exports, "logisticMF", logisticMF);
  NODE_SET_METHOD(exports, "als", als);

  // Testing Algorithms
  NODE_SET_METHOD(exports, "testLogisticMf", testLogisticMf);
  NODE_SET_METHOD(exports, "testAls", testAls);
}

NODE_MODULE(main, init)

}  // namespace demo
